import python_matlab
import impact

exclude = []
custom = {}
callback_post = {}
callback_post[r"MPC\.export"] = f"""
    name = varargin{{1}};
    build_dir = [pwd filesep '{impact.build_dir_prefix}' name '{impact.build_dir_suffix}'];
    addpath(build_dir);
    current = cd(build_dir);
    run('build.m');
    cd(current);
"""
callback_pre = {}
callback_pre[r"MPC\.export"] = f"""
    varargin = [varargin {{'context','matlab'}}];
"""
python_matlab.generate(impact,exclude=exclude,custom=custom,callback_post=callback_post,callback_pre=callback_pre)



"""
Cart-pendulum DAE
=================

Basic cart-pendulum example using differential algebraic equations
"""
from impact import *

mpc = MPC()

cart_pendulum = mpc.add_model('cart_pendulum','cart_pendulum_dae.yaml')


print(cart_pendulum.L)

# Parameters
x_current = mpc.parameter('x_current',cart_pendulum.nx)
x_final = mpc.parameter('x_final',cart_pendulum.nx)
weights = mpc.parameter('weights',2)

# Make the time horizon a parameter
T = mpc.parameter('T')
mpc.set_value(T,2.0)
mpc.set_T(T)

# Objectives
mpc.add_objective(mpc.integral(weights[0]*cart_pendulum.U**2 + weights[1]*100*cart_pendulum.X**2))

# Boundary constraints
mpc.subject_to(mpc.at_t0(cart_pendulum.x)==x_current)
mpc.subject_to(mpc.at_tf(cart_pendulum.x)==x_final)

# Path constraints
mpc.subject_to(-2 <= (cart_pendulum.U <= 2 ))
# In MPC, you typically do not want to enforce state constraints at the initial time
mpc.subject_to(-2 <= (cart_pendulum.X <= 2), include_first=False)


# Solver
#options = {"ipopt": {"print_level": 0}}
#options["expand"] = True
#options["print_time"] = False
#mpc.solver('ipopt',options)

options = {}
options["qpsol"] = "qrqp"
mpc.solver('sqpmethod',options)

mpc.set_value(x_current, [0.5,cart_pendulum.L,0, 0,0,0])
mpc.set_value(x_final, [0,cart_pendulum.L,0, 0,0,0])
mpc.set_value(weights, [1,1])


mpc.set_initial(cart_pendulum.xa, 1.11)
mpc.set_initial(cart_pendulum.x, 1.13)

# Make it concrete for this ocp
mpc.method(DirectCollocation(N=50))

# Demonstrate how to export helper functions
import casadi as cs
x = cs.MX.sym("x")
y = cs.MX.sym("y",3)
z = cs.MX.sym("z",3,3)
foo = cs.Function("foo",[x,y,z],[2*x,x*y,z@y+x],['x','y','z'],['foo','bar','baz'])
mpc.add_function(foo) # Becomes available in library*.slx

mpc.export("cart_pendulum_dae")





"""
Cart-pendulum
=============

Basic cart-pendulum example using YAML model
"""
from impact import *

mpc = MPC(T=2.0)

cart_pendulum = mpc.add_model('cart_pendulum','cart_pendulum2.yaml')

## Parameters
x_current = mpc.parameter('x_current',cart_pendulum.nx)
x_final = mpc.parameter('x_final',cart_pendulum.nx)
weights = mpc.parameter('weights',2)

## Objectives
mpc.add_objective(mpc.sum(weights[0]*cart_pendulum.F**2 + weights[1]*100*cart_pendulum.pos**2))

## Boundary constraints
mpc.subject_to(mpc.at_t0(cart_pendulum.x)==x_current)
mpc.subject_to(mpc.at_tf(cart_pendulum.x)==x_final)

## Path constraints
mpc.subject_to(-2 <= (cart_pendulum.F <= 2 ))
## In MPC, you typically do not want to enforce state constraints at the initial time
mpc.subject_to(-2 <= (cart_pendulum.pos <= 2), include_first=False)

## Solver
options = {"ipopt": {"print_level": 0}}
options["expand"] = True
options["print_time"] = False
mpc.solver('ipopt',options)

mpc.set_value(x_current, [0.5,0,0,0])
mpc.set_value(x_final, [0,0,0,0])
mpc.set_value(weights, [1,1])

## Make it concrete for this ocp
mpc.method(MultipleShooting(N=50,M=1,intg='rk'))

#mpc.method(external_method('fatrop',N=50,mode='interface',fatrop_options={"tol":1e-5}))

## Save the MPC object
# mpc.save("cart_pendulum.impact")
## Export artifacts from the MPC object
mpc.export("cart_pendulum", qp_error_on_fail=False)





"""
Generate models
===============

Basic example to generate models as CasADi functions
"""
from casadi import *

pos = MX.sym("pos")
theta = MX.sym("theta")
dpos = MX.sym("dpos")
dtheta = MX.sym("dtheta")

F = MX.sym("F")

mcart = 0.5   # cart mass [kg]
m = 1     # pendulum mass [kg]
L = 2     # pendulum length [m]
g = 9.81  # gravitation [m/s^2]

pos_rhs = dpos
theta_rhs = dtheta
dpos_rhs = (-m*L*sin(theta)*dtheta*dtheta + m*g*cos(theta)*sin(theta)+F)/(mcart + m - m*cos(theta)*cos(theta))
dtheta_rhs = (-m*L*cos(theta)*sin(theta)*dtheta*dtheta + F*cos(theta)+(mcart+m)*g*sin(theta))/(L*(mcart + m - m*cos(theta)*cos(theta)))

dae = {}
dae["x"] = vertcat(pos,theta,dpos,dtheta)
dae["u"] = F
dae["ode"] = vertcat(pos_rhs,theta_rhs,dpos_rhs,dtheta_rhs)

model = Function("model",dae,['x','u'],["ode","dae","quad"])
model.save("cart_pendulum_equations.casadi")



from ctypes import *
import numpy as np
import os

try:
    import casadi
    os.environ["PATH"] += os.pathsep + casadi.GlobalOptions.getCasadiPath()
except:
    pass

class Impact:
    def _register(self,fun_name,argtypes,restype):
        fun = getattr(self.lib,self.prefix+fun_name)
        setattr(self,"_"+fun_name,fun)
        fun.argtypes = argtypes
        fun.restype = restype

    def __init__(self, name, src_dir="."):
        self.prefix = "impact_"
        build_dir_rel = name+"_build_dir"
        build_dir_abs = os.path.join(os.path.abspath(src_dir),build_dir_rel)

        # PyDLL instead of CDLL to keep GIL:
        # virtual machine emits Python prints
        if os.name == "nt":
            libname = name+".dll"
        else:
            libname = "lib"+name+".so"
        self.lib = PyDLL(os.path.join(build_dir_abs,libname))

        # Type aliases
        m_type = c_void_p
        CONST = lambda x: x

        # Relay printing to python
        formatter_type = CFUNCTYPE(c_int, c_char_p)
        def formatter(c):
            print(c.decode('ascii'),end='')
            return 0
        self._formatter = formatter_type(formatter)

        self._register("initialize",[formatter_type, c_char_p], m_type)
        self._register("destroy",[m_type], None)
        self._register("get",[m_type, c_char_p, c_char_p, c_int, POINTER(c_double), c_int], c_int)
        self._register("set",[m_type, c_char_p, c_char_p, c_int, POINTER(CONST(c_double)), c_int], c_int)
        self._register("print_problem",[m_type], c_int)
        self._register("get_id_count",[m_type, c_char_p], c_int)
        self._register("get_size",[m_type, c_char_p, c_char_p, c_int, c_int, POINTER(c_int), POINTER(c_int)], c_int)
        self._register("solve",[m_type], c_int)
        self._register("hotstart",[m_type], c_int)
        self._register("flag_size",[m_type], c_int)
        self._register("flag_name",[m_type, c_int], c_char_p)
        self._register("flag_value",[m_type, c_int], c_int)

        self._m = self._initialize(self._formatter, build_dir_abs.encode("ascii"))
        """
          int {prefix}get_id_count(MPCstruct* m, const char* pool_name);
          int {prefix}get_id_from_index(MPCstruct* m, const char* pool_name, int index, const char** id);
          int {prefix}get_size(MPCstruct* m, const char* pool_name, const char* id, int* n_row, int* n_col);

          int {prefix}print_problem(MPCstruct* m);
          void {prefix}set_work(MPCstruct* m, const casadi_real** arg, casadi_real** res, casadi_int* iw, casadi_real* w);
          void {prefix}work(MPCstruct* m, casadi_int* sz_arg, casadi_int* sz_res, casadi_int* sz_iw, casadi_int* sz_w);

        # MPCstruct* m = initialize(printf)
        """

        for i in range(self._flag_size(self._m)):
            setattr(self, self._flag_name(self._m, i).decode("ascii"), self._flag_value(self._m, i))

    def get(self, pool_name, id, stage, flags):
        shape = self.get_size(pool_name, id, stage, flags)
        ret = np.zeros(shape,dtype=np.float64)
        if 1 not in shape:
            flags |= self.ROW_MAJOR
        self._get(self._m, pool_name.encode("ascii"), None if id==self.ALL else id.encode("ascii"), stage, ret.ctypes.data_as(POINTER(c_double)), flags)
        return ret

    def set(self, pool_name, id, stage, flags, data):
        if isinstance(data,int) or  isinstance(data,float):
          data = np.array([data],dtype=float)
        elif isinstance(data,list):
          data = np.array(data,dtype=float)
        shape = self.get_size(pool_name, id, stage, flags)
        if len(data.shape)==1 and 1 in shape:
            pass
        elif data.shape==shape:
            flags |= self.ROW_MAJOR
        else:
            raise Exception(f"Expected shape {shape}, got {data.shape} instead.")
        self._set(self._m, pool_name.encode("ascii"), None if id==self.ALL else id.encode("ascii"), stage, data.ctypes.data_as(POINTER(c_double)), flags)

    def get_size(self, pool_name, id, stage, flags):
        row = c_int()
        col = c_int()
        self._get_size(self._m, pool_name.encode("ascii"), None if id==self.ALL else id.encode("ascii"), stage, flags, byref(row), byref(col))
        return (row.value, col.value)

    def print_problem(self):
        self._print_problem(self._m)

    def solve(self):
        self._solve(self._m)

    def hotstart(self):
        self._hotstart(self._m)

    def __del__(self):
        if hasattr(self, "_m"):
            self._destroy(self._m)



"""
Cart-pendulum features
======================

Basic cart-pendulum example
"""
from impact import *

from rockit import *

from casadi import sumsqr

mpc = MPC(T=FreeTime(2.0))

cart_pendulum = mpc.add_model('cart_pendulum','cart_pendulum2.yaml')

# Parameters
x_current = mpc.parameter('x_current',cart_pendulum.nx)
x_final = mpc.parameter('x_final',cart_pendulum.nx)
weights = mpc.parameter('weights',2)

#w = mpc.parameter('w',cart_pendulum.nx,grid='control')

#v = mpc.variable("v")

#mpc.set_initial(v0,42)

#w = mpc.parameter('w',cart_pendulum.nx,grid='control')

#mpc.set_initial(v0,42)

print(mpc.variables)

# Objectives
mpc.add_objective(mpc.integral(weights[0]*cart_pendulum.F**2 + weights[1]*100*cart_pendulum.pos**2))

# Boundary constraints
mpc.subject_to(mpc.at_t0(cart_pendulum.x)==x_current)
mpc.subject_to(mpc.at_tf(cart_pendulum.x)==x_final)

# Path constraints
mpc.subject_to(-2 <= (cart_pendulum.F <= 2 ))
# In MPC, you typically do not want to enforce state constraints at the initial time
mpc.subject_to(-2 <= (cart_pendulum.pos <= 2), include_first=False)



# Solver
options = {"ipopt": {"print_level": 0}}
options["expand"] = True
options["print_time"] = False
mpc.solver('ipopt',options)

mpc.set_value(x_current, [0.5,0,0,0])
mpc.set_value(x_final, [0,0,0,0])
mpc.set_value(weights, [1,1])

#mpc.set_value(w, 3)

# Make it concrete for this ocp
mpc.method(MultipleShooting(N=50,M=1,intg='rk'))

mpc.save("cart_pendulum.impact")



mpc.export("cart_pendulum",qp_error_on_fail=False)


mpc2 = MPC.load("cart_pendulum.impact")
mpc2.export("cart_pendulum2",qp_error_on_fail=False)



import os



def simulink_parse_states(mdl,unzipped_path):

    def get_full_id(id):
        with open(os.path.join(unzipped_path,'sources',id+"_gateway.c"),'r') as inp:
            for line in inp:
                line = line.strip()
                if "_".join(id.split("_")[:-1]) in line:
                    return line.split('"')[1][:-2]

    def get_state_names(full_id):
        with open(os.path.join(unzipped_path,'sources',full_id+"_ds.c"),'r') as inp:
            harvesting = False
            harvest = ""
            for line in inp:
                line = line.strip()
                if "NeVariableData s_variable_data" in line:
                    harvesting = True
                    harvest = line
                elif harvesting:
                    harvest += line
                    if ";" in line:
                        break



        harvest = harvest[harvest.index("{"):harvest.rindex("}")+1]

        from pyparsing import nestedExpr, delimitedList
        content = nestedExpr(opener="{", closer="}", content=delimitedList(nestedExpr(opener="{", closer="}")))
        harvest = content.parseString(harvest,parseAll=True).asList()[0]
        names = []
        for e in harvest:
            names.append(e[0][1:-1])

        return names

    state_class = 'X_'+mdl+'_T'

    typedefs = {}

    typedef_staging = None
    prev_line = None
    simscape_states = []
    simscape_ids = []

    with open(os.path.join(unzipped_path,'sources',mdl+'.c'),'r') as inp:
        for line in inp:
            line = line.strip()
            if "_gateway(" in line:
                active_simscape = line[:line.index("(")].rstrip()
                if active_simscape not in simscape_ids:
                    simscape_ids.append(active_simscape[:-len("_gateway")])
            if "simulationData->mData->mContStates.mX" in line and "NULL" not in line:
                n = int(prev_line.split("=")[1][:-1])
                name = line.split("->")[-1].split("[")[0]
                if (name,n) not in simscape_states:
                    simscape_states.append((name,n))
            prev_line = line

    simscape_state = {}

    for k in range(len(simscape_states)):
        short_id = simscape_ids[k]
        full_id = get_full_id(short_id)
        names = get_state_names(full_id)
        assert len(names)==simscape_states[k][1]
        simscape_state[simscape_states[k][0]] = names

    print(simscape_state)

    with open(os.path.join(unzipped_path,'sources',mdl+'.h'),'r') as inp:
        for line in inp:
            line = line.strip()
            if "typedef struct" in line:
                typedef_staging = []
            else:
                if typedef_staging is not None:
                    if "}" in line:
                        typedefs[line[1:-1].strip()] = typedef_staging
                        typedef_staging = None
                    else:
                        line = line[:-1]
                        [t,name] = line.split(" ")
                        typedef_staging.append((t,name))
    
    states = typedefs[state_class]

    for e in states:
        assert e[0]=='real_T'

    states = [e[1] for e in states]

    all_states = []
    for e in states:
        if "[" in e:
            name = e[:e.index("[")]
            index = int(e[e.index("[")+1:e.index("]")])
            assert name in simscape_state
            assert index==len(simscape_state[name])
            all_states+=simscape_state[name]
        else:
            all_states.append(e)

    return all_states